/******************************************************************************
 *
 * Filename: cycdefs.h
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  5th June 2003
 *
 * Purpose:  Parse tree structure definitions.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/
       
#ifndef __cycdefs_h

#define __cycdefs_h


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

void yyerror( char* s );


#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


/*============================================================================
 ||                  Structures which form parse tree                       ||
 ============================================================================*/

/*****************************************************************************
 * Line Number
 *****************************************************************************/

typedef struct LineNumber
{
	const char		*filename;
	int				 line;
} LineNumber;

LineNumber* newLineNumber( const char* filename, int line );

LineNumber* copyLineNumber( LineNumber *lineNo );

void freeLineNumber( LineNumber *lineNo );

/*****************************************************************************
 * Identifer
 *****************************************************************************/

typedef enum { untyped_id=0, integer_id, string_id } IdentType;

typedef struct
{
	const char            *name;
	IdentType              type;
	LineNumber 			  *lineNo;
} Ident;

Ident* newIdent( char* n, IdentType t );

Ident* copyIdent( Ident *ident );

void freeIdent( Ident *ident );

typedef struct IdentList
{
	Ident            *head;
	struct IdentList *tail;
} IdentList;

IdentList *newIdentList( Ident *el, IdentList *l );

void freeIdentList( IdentList *idlist );


/*****************************************************************************
 * Class
 *****************************************************************************/

typedef struct
{
	Ident			  *name;
	IdentList		  *pArgs;
	struct MemberList *pMembers;

	/* Semantic information */
	struct _Hashtable *memberTable;		/* Map between member name and Member */
} Class;

Class* newClass( Ident* name, IdentList *pArgs, struct MemberList *pMembers );

void freeClass( Class *clazz );

typedef struct ClassList
{
	Class            *head;
	struct ClassList *tail;
} ClassList;

ClassList* newClassList( Class *el, ClassList *l );


/*****************************************************************************
 * Procedure or Function
 *****************************************************************************/

typedef struct
{
	Ident                *name; 
	struct IdentList     *pArgs;
	struct StatementList *pStmts;
} Procedure;

Procedure* newProcedure( Ident* n, IdentList* pa, struct StatementList *ps );

typedef struct ProcedureList
{
	Procedure            *head;
	struct ProcedureList *tail;
} ProcedureList;

ProcedureList* newProcedureList( Procedure *el, ProcedureList *l );


/******************************************************************************
 * Various forms of Member
 *****************************************************************************/

typedef enum
{
	block_def
} MemberDefType;

typedef struct
{
	MemberDefType  kind;
	struct ExpressionList *pArgs;
} MemberDef;

MemberDef* newMemberDef( MemberDefType kind, struct ExpressionList *pArgs );

typedef enum
{
	const_mbr, input_mbr, process_mbr, output_mbr
} MemberType;

typedef struct
{
	struct Expression *from;
	struct Expression *to;
} Range;

Range* newRange( struct Expression *from, struct Expression *to );

typedef struct
{
	MemberType	 kind;
	Ident		*name;
	union
	{
		struct
		{
			struct Expression *value;
		} const_part;
		struct
		{
			IdentList	*pArgs;
			MemberDef   *definition;
		} normal_part;
		struct
		{
			IdentList	*pArgs;
			MemberDef   *definition;
			Range		*range;
		} input_part;
	} data;

	/* Semantic information */
	Class		*parent;		/* Parent class */
} Member;

Member* newConstMember( Ident *name, struct Expression *value );

Member* newMember( MemberType kind, Ident *name, IdentList *pArgs, MemberDef *def );

Member* newInputMember( Ident *name, IdentList *pArgs, MemberDef *def, Range *range );

void freeMember( Member *member );

typedef struct MemberList
{
	Member            *head;
	struct MemberList *tail;
} MemberList;

MemberList* newMemberList( Member *el, MemberList *l );

void freeMemberList( MemberList *mbrlist );


/******************************************************************************
 * Instances
 *****************************************************************************/

typedef struct
{
	Ident *name;
	Ident *type;
	struct ExpressionList *pArgs;

	/* Semantic information */
	Class *clazz;	/* Actual Class representing type */
} Instance;

Instance *newInstance( Ident *type, Ident *name, struct ExpressionList *pArgs );

typedef struct InstanceList
{
	Instance            *head;
	struct InstanceList *tail;
} InstanceList;

InstanceList* newInstanceList( Instance *el, InstanceList *l );


/******************************************************************************
 * Various forms of Statement
 *****************************************************************************/

typedef enum
{
	/* Real statements */
	if_stmt, while_stmt, do_stmt, for_stmt, switch_stmt,
	assignment_stmt, call_stmt, return_stmt, break_stmt,
	/* Pseudo-statements*/
	start_stmt, goto_stmt, label_stmt
} StmtType;

typedef enum
{
	equals_assign, plusequals_assign, minusequals_assign, timesequals_assign,
	divideequals_assign, increment_assign, decrement_assign
} AssignmentType;

typedef struct Statement
{
	StmtType		  kind;
	LineNumber		 *lineNo;

	union
	{
		struct
		{
			struct Expression      *condition;
			struct StatementList   *pThen;
			struct StatementList   *pElse;
		} if_part;      
		struct
		{
			struct Expression      *condition;
			struct StatementList   *pBody;
		} while_and_do_part;      
		struct
		{
			struct Statement	   *initial;
			struct Expression	   *condition;
			struct Statement	   *increment;
			struct StatementList   *pBody;
		} for_part;
		struct
		{
			struct Expression	   *expr;
			struct CaseBlockList   *pCaseBlocks;
		} switch_part;
		struct
		{
			AssignmentType		   kind;
			Ident                  *name;
			struct Expression      *value;
		} assignment_part;      
		struct
		{
			Ident                 *clazz;
			Ident                 *name;
			struct ExpressionList *pArgs;

			/* Semantic information */
			union
			{
				Procedure *proc;	/* The Procedure this call represents */
				Member	  *member;	/* The class Member this call represents */
			} definition;
		} call_part;
		struct
		{
			struct Expression     *expr;
		} goto_and_label_part;
	} data;
} Statement;

typedef struct CaseBlock
{
	struct ExpressionList		  *pCases;
	struct StatementList		  *pStmts;
} CaseBlock;

typedef struct CaseBlockList
{
	CaseBlock			 *head;
	struct CaseBlockList *tail;
} CaseBlockList;

typedef struct StatementList
{
	Statement            *head;
	struct StatementList *tail;
} StatementList;

Statement* newIfStatement( struct Expression *cond, StatementList *pThen, StatementList *pElse );

Statement* newWhileStatement( struct Expression *cond, StatementList *pBody );

Statement* newDoStatement( struct Expression *cond, StatementList *pBody );

Statement* newForStatement( Statement *init, struct Expression *cond, Statement *incr, StatementList *pBody );

Statement* newSwitchStatement( struct Expression *expr, CaseBlockList *pCaseBlocks );

CaseBlock* newCaseBlock( struct ExpressionList *pCases, StatementList *pStmts );

CaseBlockList *newCaseBlockList( CaseBlock *head, CaseBlockList *tail );

Statement* newAssignmentStatement( Ident *name, AssignmentType kind, struct Expression *value );

Statement* newCallStatement( Ident* clazz, Ident *name, struct ExpressionList *pArgs );

Statement* newGotoStatement( struct Expression *expr );

Statement* newLabelStatement( struct Expression *expr );

Statement* newSimpleStatement( int kind );

StatementList *newStatementList( Statement *el, StatementList *l );

void freeStatement( Statement* stmt );

void freeStatementList( StatementList* stmtlist );


/*****************************************************************************
 * Various forms of Expression
 *****************************************************************************/

typedef enum
{
	variable_expr, integer_expr, string_expr, boolean_expr,
	addition_expr, subtraction_expr, multiplication_expr,
	division_expr, equality_expr, inequality_expr,
	lessthan_expr, greaterthan_expr, lessequal_expr,
	greaterequal_expr, call_expr, constant_expr, default_expr
} ExprKind;

typedef enum
{
	untyped_type=untyped_id, integer_type=integer_id,
	string_type=string_id, boolean_type
} ExprType;

typedef struct Expression
{
	ExprKind     kind;
	ExprType     type;
	LineNumber	*lineNo;

	union
	{
		Ident             *variable_part;
		int                integer_part;
		const char        *string_part;
		struct
		{
			struct Expression      *left;
			struct Expression      *right;
		} binary_op_part;
		struct
		{
			Ident					*clazz;
			Ident					*name;
			struct ExpressionList	*pArgs;

			/* Semantic information */
			Member					*definition;	/* The class Member this call represents */
		} call_part;
		struct
		{
			Ident					*clazz;
			Ident					*name;

			/* Semantic information */
			struct Expression		*value;			/* The resolved value of this constant */
		} constant_part;
		struct
		{
			/* Semantic information */
			struct ExpressionList	*pCases;
		} default_part;
	} data;
} Expression;

Expression* newVariableExpr( Ident *pvar );
Expression* newIntegerExpr( int k, int n );
Expression* newStringExpr( const char *s );

Expression* newBinaryOpExpression( ExprKind k, Expression* l, Expression* r );
Expression* newCallExpression( Ident* clazz, Ident* name, struct ExpressionList* pArgs );
Expression* newConstantExpression( Ident* clazz, Ident* name );
Expression* newDefaultExpr();

void freeExpression( Expression *expr );

typedef struct ExpressionList
{
	Expression            *head;
	struct ExpressionList *tail;
} ExpressionList;

ExpressionList *newExpressionList( Expression *el, ExpressionList *l );

void freeExpressionList( ExpressionList *exprlist );

const char* typeToString( ExprType t );


/*****************************************************************************
 * Program
 *****************************************************************************/

typedef enum
{
	class_pic, instance_pic, proc_pic
} PICType;

typedef struct
{
	PICType kind;
	union
	{
		Class	  *clazz;
		Instance  *instance;
		Procedure *proc;
	} data;
} PIC;

PIC *newProcedurePIC( Procedure *proc );
PIC *newInstancePIC( Instance *instance );
PIC *newClassPIC( Class *clazz );

typedef struct PICList
{
	PIC            *head;
	struct PICList *tail;
} PICList;

PICList *newPICList( PIC *el, PICList *l );

typedef struct
{
	ClassList	    *pClasses;
	InstanceList    *pInstances;
	ProcedureList   *pProcs;
	PICList		    *pPICs;
	struct LineList *pLines;
	const char *sourcefile;
	const char *outputfile; 
} Program;

Program* newProgram( ProcedureList *pp );


/*===========================================================================*/

/* Code generation stuff */

typedef struct
{
	int fromid;
	int fromport;
	int toid;
	int toport;
} Line;

Line* newLine( int fid, int fport, int tid, int tport );

typedef struct LineList
{
	Line            *head;
	struct LineList *tail;
} LineList;

LineList *newLineList( Line *el, LineList *l );


extern Program prog;

extern struct _Hashtable* classTable;

extern void cleanup( Program *prog );


#ifdef __cplusplus
}
#endif

#endif /* __cycdefs_h */
