/******************************************************************************
 *
 * Filename: cycle.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  5th June 2003
 *
 * Purpose:  The compiler main program.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(WIN32) && CRTMEMCHECK
#include <crtdbg.h>
_CrtMemState crts1, crts2, crts3;
#endif /* WIN32 && CRTMEMCHECK */

#include "cycdefs.h"
#include "cycpars.h"
#include "errors.h"

#include "getopt.h"

#ifdef WIN32
#include "windows.h"
#include "winreg.h"
#define strdup _strdup
#endif /* WIN32 */

extern int yyparse(void);
extern void lexer_cleanup();
extern FILE *yyin;
extern int yydebug;
extern int yy_flex_debug;
int debug_level = 0;
int dump_tree	= 0;
int verbose        = FALSE;
int translate_only = FALSE;
int ignore_errors  = FALSE;
int no_logo        = FALSE;
#ifdef WIN32
int install        = FALSE;
#endif /* WIN32 */
int ret;
int errors, warnings;

const char* infile  = NULL;
const char* exepath = NULL;

int option;
char* optarg;
int optind;

Program prog;

void semanticCheck( Program* );
void transformCode( Program* );
void dumpCode( Program* );
void generateCode( Program* );


void usage()
{
	fprintf( stderr, VERSION "\n" );
	fprintf( stderr, COPYRIGHT "\n\n" );
#ifdef WIN32
	fprintf( stderr, "Usage: " PROGNAME " [-ehinptvV] [-d <n>] [-o <file>] <file>\n" );
#else
	fprintf( stderr, "Usage: " PROGNAME " [-ehnptvV] [-d <n>] [-o <file>] <file>\n" );
#endif /* WIN32 */
	fprintf( stderr, "      -d <n>    = debug 1 - parser, 2 - lexer, 3 - both\n" );
	fprintf( stderr, "      -e        = ignore errors (only use with -p and -t)\n" );
	fprintf( stderr, "      -h        = this message\n" );
#ifdef WIN32
	fprintf( stderr, "      -i        = install output file in RR folder\n" );
#endif /* WIN32 */
	fprintf( stderr, "      -n        = no logo\n" );
	fprintf( stderr, "      -o <file> = output file\n" );
	fprintf( stderr, "      -p        = show parse tree\n" );
	fprintf( stderr, "      -t        = translate only\n" );
	fprintf( stderr, "      -v        = verbose\n" );
	fprintf( stderr, "      -V        = show version information\n" );
}


int main( int argc, char *argv[] )
{
	int len;

#if defined(WIN32) && CRTMEMCHECK
    _CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDOUT );
	_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDOUT );
	_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDOUT );

   _CrtMemCheckpoint( &crts1 );
#endif /* WIN32 && CRTMEMCHECK */

	yyin    = NULL;
	yydebug = FALSE;
	yy_flex_debug = FALSE;
	errors = warnings = 0;
	prog.outputfile = strdup( "output.03p" );

	if( argc == 1 )
	{
		usage();
		return 1;
	}

	/* Check to see if the program was run with a path in front of it */
	if( (len = strlen( argv[0] )) > 9 )
	{
		if( stricmp( &argv[0][len - 9], "cycle.exe" ) == 0 )
		{
			char* path = strdup( argv[0] );
			path[len-9] = '\0';
			exepath = path;
		}
		else if( stricmp( &argv[0][len - 5], "cycle" ) == 0 )
		{
			char* path = strdup( argv[0] );
			path[len-5] = '\0';
			exepath = path;
		}
		else
			exepath = strdup( "." );
	}
	else
		exepath = strdup( "." );

#ifdef WIN32
	optind = 1;

	while( (optind < argc) && (argv[optind][0] == '-') )
	{
		option = argv[optind][1];

		if( option != '\0' )
			optarg = &argv[optind][2];

		/* Check for stray arguments */
		if( (option != 'd') && (option != 'o') && (optarg[0] != '\0') )
		{
			usage();
			return 1;
		}
#else
	while( (option = getopt( argc, argv, "cd:eghino:pvV" )) != EOF )
	{
#endif

		switch( option )
		{
		case '\0':
			infile = strdup( optarg );
			break;
		case 'V':
			return 0;
		case 'd':
			debug_level = atoi( optarg );
			yydebug       = ( debug_level & 1 ? TRUE : FALSE );
			yy_flex_debug = ( debug_level & 2 ? TRUE : FALSE );
			break;
		case 'e':
			ignore_errors = TRUE;
			break;
#ifdef WIN32
		case 'i':
			install = TRUE;
			break;
#endif /* WIN32 */
		case 'n':
			no_logo = TRUE;
			break;
		case 'o':
			free( (char*) prog.outputfile );
			prog.outputfile = strdup( optarg );
			break;
		case 'p':
			dump_tree++;
			break;
		case 't':
			translate_only = TRUE;
			break;
		case 'v':
			verbose = TRUE;
			break;
		case 'h':
		default:
			usage();
			return 1;
		}

#ifdef WIN32
		optind++;
#endif
	}

	if( (optind < argc) && !infile )
		infile = strdup( argv[optind] );

	if( !infile )
	{
		fprintf( stderr, VERSION "\n" );
		fprintf( stderr, COPYRIGHT "\n" );
		fprintf( stderr, PROGNAME ": No input file specified\n" );
		return 1;
	}

	if( !no_logo )
	{
		fprintf( stderr, VERSION "\n" );
		fprintf( stderr, COPYRIGHT "\n" );
	}

	prog.pProcs = NULL;

	/* shorten sourcefile to just leafname */
	{
		int start, end;

		for( end = strlen( infile ); (end > 0) && (infile[end] != '.'); end-- );
		end--; /* before '.' */
		for( start = end-1; (start > 0) && (infile[start] != '.') && (infile[start] != '/') && (infile[start] != '\\'); start-- );
		start++; /* after sep */

		prog.sourcefile = strdup( &infile[start] );
	}

	yyin = fopen( infile, "r" );

	if( !yyin )
	{
		fprintf( stderr, PROGNAME ": unable to open input file, \"%s\"\n", infile );
		return 1;
	}

	verbose_msg( infile, "Parsing..." );

	ret = yyparse();

	fclose( yyin );

	/* Clean up lexer buffer */
	lexer_cleanup();

	if( !ignore_errors && (ret || errors) )
	{
		cleanup( &prog );
		return 2;
	}

	verbose_msg( infile, "syntax OK" );

	verbose_msg( infile, "Checking semantics..." );
    semanticCheck( &prog );

	if( !ignore_errors && errors )
	{
		cleanup( &prog );
		return 3;
	}

	if( dump_tree & 1 )
		dumpCode( &prog );

	verbose_msg( infile, "Transforming code..." );
	transformCode( &prog );

	if( !ignore_errors && errors )
	{
		cleanup( &prog );
		return 4;
	}

	if( dump_tree & 2 )
		dumpCode( &prog );

	if( !translate_only )
	{
#ifdef WIN32
		if( install )
		{
			HKEY hKey;
			DWORD valueType, valueSize;
			BYTE valueData[256];

			if( (RegOpenKeyEx( HKEY_LOCAL_MACHINE, "SOFTWARE\\Maris Technologies\\Real Robots2\\Settings", 0, KEY_READ, &hKey ) == ERROR_SUCCESS)
				&& (RegQueryValueEx( hKey, "WorkDir", NULL, &valueType, valueData, &valueSize ) == ERROR_SUCCESS)
				&& (RegCloseKey( hKey ) == ERROR_SUCCESS) )
			{
				char buf[256];
				valueData[valueSize] = '\0';
				sprintf( buf, "%s\\program files\\Real Robots\\Robot Programs\\%s", valueData, prog.outputfile );
				free( (char*) prog.outputfile );
				prog.outputfile = strdup( buf );
			}
		}
#endif /* WIN32 */

		verbose_msg1( infile, "Generating code to '%s'...", prog.outputfile );

		generateCode( &prog );
	}

	cleanup( &prog );

	return ignore_errors ? 0 : (errors ? 5 : 0);
}
