/******************************************************************************
 *
 * Filename: cyclex.l
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  5th June 2003
 *
 * Purpose:  Flex scanner for Cycle language.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

%{
#include <string.h>
#include <stdlib.h>

#include "cycdefs.h"
#include "cycpars.h"
#include "errors.h"

#ifdef WIN32
#define strdup _strdup
#endif /* WIN32 */

extern const char* exepath;
FILE* inclin = NULL;
YY_BUFFER_STATE included = 0;
YY_BUFFER_STATE original = 0;
const char* origfile = NULL;
int origLineNo = 0;

/* global line count */
int _lineNo = 1;
%}

identifier      [A-Za-z_][A-Za-z0-9_]*
      
%x cmnt
%x strcon

%%
                char string_buf[255];
                char *string_buf_ptr;

[ \t]+          /* eat up whitespace */

\n              ++_lineNo;

"//"[^\n]*      /* eat up C++ comments */

"/*"	        { /* eat up C comments */
                  BEGIN(cmnt);
                }
                
<cmnt>[^\n]*"*/" { /* end of C comment */
                  BEGIN(INITIAL);
                }

<cmnt>[^\n]+	/* body of C comment */

<cmnt>\n		{ /* body of C comment */
				  ++_lineNo;
				}

\"              {
                  string_buf_ptr = string_buf;
                  BEGIN(strcon);
                }

<strcon>\"      { /* saw closing quote - all done */
                  BEGIN(INITIAL);
                  *string_buf_ptr = '\0';
                  yylval.string = strdup( string_buf );
                  return T_STRING;
                }

<strcon>\n      yyerror( "unterminated string constant" );

<strcon>[^\n\"] { /* + */
                  *string_buf_ptr++ = yytext[0];
                  /* char *yytext_ptr = yytext;
                  while ( *yytext_ptr )
                    *string_buf_ptr++ = *yytext_ptr++; */
                }

block			return T_BLOCK;
break           return T_BREAK;
case            return T_CASE;
class           return T_CLASS;
const			return T_CONST;
default			return T_DEFAULT;
do              return T_DO;
else            return T_ELSE;
for             return T_FOR;
if              return T_IF;
include         return T_INCLUDE;
input			return T_INPUT;
output			return T_OUTPUT;
proc            return T_PROC;
process			return T_PROCESS;
return          return T_RETURN;
switch          return T_SWITCH;
while           return T_WHILE;

"++"			return T_INCR;
"--"			return T_DECR;

"+="			return T_PLUSEQ;
"-="			return T_MINUSEQ;
"*="			return T_TIMESEQ;
"/="			return T_DIVIDEEQ;

"=="            return T_EQUALS;
"<="            return T_LEQ;
">="            return T_GEQ;
"!="            return T_NEQ;

false           {
                  yylval.integer = 0;
                  return T_BOOLEAN;
                }
                
true            {
                  yylval.integer = 1;
                  return T_BOOLEAN;
                }

{identifier}    {
                  yylval.id = newIdent( strdup( yytext ), untyped_id );
                  return T_IDENTIFIER;
                }

"-"[0-9]+       {
                  yylval.integer = atoi( yytext );
                  return T_INTEGER;
                }

[0-9]+          {
                  yylval.integer = atoi( yytext );
                  return T_INTEGER;
                }

.               return (int) yytext[0];

%%

extern const char *infile;
extern int errors;

/**
 * Called by yyparse on error
 */
void yyerror( char *s )
{
	fprintf( stderr, "\"%s\", line %d: %s with '%s'\n", infile, _lineNo, s, yytext);
	errors++;
}


/**
 * Called at the end of a file to get next one ready.  Used to revert to the
 * main buffer when include file is finished.
 */
int yywrap()
{
	if( included )
	{
		verbose_msg( infile, "Syntax OK" );

		/* Restore the old state */
		free( (char*) infile );
		infile = origfile;
		origfile = NULL;
		_lineNo = origLineNo;

		/* Do the switch */
		yy_switch_to_buffer( original );

		/* Clean up */
		yy_delete_buffer( included );
		fclose( inclin );
		original = included = 0;
		inclin = NULL;
		return 0;
	}

	return 1;
}


/**
 * Called to parse an included file.
 */
void parse( const char* file )
{
	/* Make sure we only include to one level deep */
	if( included )
	{
		fprintf( stderr, "error: attempt to nest include files\n" );
		free( (char*) file );
		return;
	}

	/* Try current directory */
	inclin = fopen( file, "r" );
  
	if( !inclin )
	{
		/* Try relative to install dir */
		char buf[256];
		sprintf( buf, "%s..\\include\\%s", exepath, file );
		inclin = fopen( buf, "r" );

		if( !inclin )
		{
			fprintf( stderr, "error: unable to open input file, \"%s\"\n", file );
			return;
		}
	}

	verbose_msg( file, "Parsing..." );

	/* Remember current state */
	origfile = infile;
	origLineNo = _lineNo;
	original = YY_CURRENT_BUFFER;

	/* Load new state */
	infile = file;
	_lineNo = 1;
	included = yy_new_buffer( inclin, 8096 );

	/* Do the switch */
	yy_switch_to_buffer( included );
}


void lexer_cleanup()
{
	if( included )
		yywrap();

	yy_delete_buffer( YY_CURRENT_BUFFER );
}
