/******************************************************************************
 *
 * Filename: cycpars.y
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  5th June 2003
 *
 * Purpose:  Bison grammar for Cycle language.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

%{                       /* C declarations */
#include "cycdefs.h"
#include "errors.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef __GNUC__
#define alloca malloc
#endif

#if defined(WIN32) || defined(__GNUC__)
#define xmalloc malloc
#endif

#define YYDEBUG 1
#define YYERROR_VERBOSE 1
extern int yylex( void );
void yyerror( char* s );
//extern int _lineNo;
extern void parse( const char* file );
%}

/*%pure_parser*/

%union
{
  char           *string;
  int             integer;
  Ident          *id;
  IdentList      *idlist;
  CaseBlock      *caseblk;
  CaseBlockList  *cblist;
  Statement      *stmt;
  StatementList  *slist;
  Expression     *expr;
  ExpressionList *elist;
  MemberList	 *mbrlist;
  Member		 *member;
  MemberDef		 *mbrdef;
  Range			 *range;
  PIC			 *pic;
  Program        *prog;
}

/* Parsable tokens */

%token <integer>  T_INTEGER T_BOOLEAN
%token <id>       T_IDENTIFIER
%token <string>   T_STRING

%token T_BLOCK T_BREAK T_CASE T_CLASS T_CONST T_DEFAULT T_DO T_ELSE T_FOR
%token T_IF T_INCLUDE T_INPUT T_OUTPUT T_PROC T_PROCESS T_RETURN T_SWITCH
%token T_WHILE

%token T_INCR T_DECR

%token T_PLUSEQ T_MINUSEQ T_TIMESEQ T_DIVIDEEQ

%token T_EQUALS T_LEQ T_GEQ T_NEQ


/* Return Types of the non-terminals */

%type <prog>    program
%type <pic>		declaration
%type <mbrlist> opt_member_decl_list
%type <member>  member_decl
%type <mbrdef>  semicolon_or_definition
%type <range>	opt_range
%type <idlist>  opt_argument_declaration_list  argument_declaration_list  opt_argument_declaration_list_tail
%type <slist>   opt_statement_list  statement_block  opt_else_clause
%type <stmt>    statement  assignment_statement
%type <caseblk> case_block
%type <cblist>  opt_case_block_list
%type <elist>   opt_case_clause_list  opt_expression_list
%type <elist>   expression_list  expression_list_tail
%type <expr>    case_clause  comparison_expression
%type <expr>    primary_expression  additive_expression

/* Grammar follows */

%%

/*******************************************************************************
 *                               Program                                       *
 ******************************************************************************/

program:
	/* empty */
	  { $$ = &prog; }
  | declaration program
      {
		  $$ = $2;

		  if( $1 )
		  {
			  switch( $1->kind )
			  {
			  case proc_pic:
				  $$->pProcs = newProcedureList( $1->data.proc, $2->pProcs );
				  break;
			  case instance_pic:
				  $$->pInstances = newInstanceList( $1->data.instance, $2->pInstances );
				  break;
			  case class_pic:
				  $$->pClasses = newClassList( $1->data.clazz, $2->pClasses );
				  break;
			  }

			  $$->pPICs = newPICList( $1, $2->pPICs );
		  }
	  }
  ;

declaration:
    T_INCLUDE T_STRING ';'
	  { parse( $2 ); $$ = NULL; }
  | T_INCLUDE error ';'
      { syntax_error( "bad include statement." ); $$ = NULL; }
  | T_CLASS T_IDENTIFIER '{' opt_member_decl_list '}'
      { $$ = newClassPIC( newClass( $2, NULL, $4 ) ); }
  | T_CLASS T_IDENTIFIER '(' argument_declaration_list ')' '{' opt_member_decl_list '}'
      { $$ = newClassPIC( newClass( $2, $4, $7 ) ); }
  | T_IDENTIFIER T_IDENTIFIER ';'
      { $$ = newInstancePIC( newInstance( $1, $2, NULL ) ); }
  | T_IDENTIFIER T_IDENTIFIER '(' expression_list ')' ';'
      { $$ = newInstancePIC( newInstance( $1, $2, $4 ) ); }
  | T_IDENTIFIER T_IDENTIFIER error ';'
      { syntax_error( "bad instance declaration." ); freeIdent( $1 ); freeIdent( $2); $$ = NULL; }
  | T_PROC T_IDENTIFIER '(' opt_argument_declaration_list ')' '{' opt_statement_list '}'
      { $$ = newProcedurePIC( newProcedure( $2, $4, $7 ) ); }
  ;


/*******************************************************************************
 *                               Classes                                       *
 ******************************************************************************/

opt_member_decl_list:
    /* empty */
	  { $$ = NULL; }
  | member_decl opt_member_decl_list
      { $$ = newMemberList( $1, $2 ); }
  ;

member_decl:
    T_CONST T_IDENTIFIER '=' primary_expression ';'
	  { $$ = newConstMember( $2, $4 ); }
  | T_INPUT opt_range T_IDENTIFIER '(' opt_argument_declaration_list ')' semicolon_or_definition
	  { $$ = newInputMember( $3, $5, $7, $2 ); }
  | T_PROCESS T_IDENTIFIER '(' opt_argument_declaration_list ')' semicolon_or_definition
	  { $$ = newMember( process_mbr, $2, $4, $6 ); }
  | T_OUTPUT T_IDENTIFIER '(' opt_argument_declaration_list ')' semicolon_or_definition
	  { $$ = newMember( output_mbr, $2, $4, $6 ); }
  ;

semicolon_or_definition:
    ';'
	  { $$ = NULL; }
  | ':' T_BLOCK '(' opt_expression_list ')' ';'
	  { $$ = newMemberDef( block_def, $4 ); }

opt_range:
    /* empty */
      { $$ = NULL; }
  | '{' primary_expression ',' primary_expression  '}'
      { $$ = newRange( $2, $4 ); }
  ;

/*******************************************************************************
 *                             Procedures                                      *
 ******************************************************************************/

opt_argument_declaration_list:
    /* empty */
      { $$ = NULL; }
  | argument_declaration_list
      { $$ = $1; }
  ;

argument_declaration_list:
    T_IDENTIFIER opt_argument_declaration_list_tail
      { $$ = newIdentList( $1, $2 ); }
 ;

opt_argument_declaration_list_tail:
    /* empty */
      { $$ = NULL; }
  | ',' T_IDENTIFIER opt_argument_declaration_list_tail
      { $$ = newIdentList( $2, $3 ); }
  ;

/*******************************************************************************
 *                             Statements                                      *
 ******************************************************************************/

opt_statement_list:
    /* empty */
      { $$ = NULL; }
  | statement opt_statement_list
      { $$ = newStatementList( $1, $2 ); }
  ;

statement:
    T_IF '(' comparison_expression ')' statement_block opt_else_clause
      { $$ = newIfStatement( $3, $5, $6 ); }
  | T_WHILE '(' comparison_expression ')' statement_block
      { $$ = newWhileStatement( $3, $5 ); }
  | T_DO statement_block T_WHILE '(' comparison_expression ')' ';'
      { $$ = newDoStatement( $5, $2 ); }
  | T_FOR '(' T_IDENTIFIER '=' additive_expression
		  ';' comparison_expression
		  ';' assignment_statement ')' statement_block
      { $$ = newForStatement( newAssignmentStatement( $3, equals_assign, $5 ), $7, $9, $11 ); }
  | T_SWITCH '(' primary_expression ')' '{' case_block opt_case_block_list '}'
      { $$ = newSwitchStatement( $3, newCaseBlockList( $6, $7 ) ); }
  | assignment_statement ';'
      { $$ = $1; }
  | T_RETURN ';'
      { $$ = newSimpleStatement( return_stmt ); }
  | T_IDENTIFIER '(' opt_expression_list ')' ';'
      { $$ = newCallStatement( NULL, $1, $3 ); }
  | T_IDENTIFIER '.' T_IDENTIFIER '(' opt_expression_list ')' ';'
      { $$ = newCallStatement( $1, $3, $5 ); }
  | T_BREAK ';'
      { $$ = newSimpleStatement( break_stmt ); }
  | error ';'
	  { syntax_error( "bad statement." ); $$ = NULL; }
  ;

statement_block:
    '{' opt_statement_list '}'
      { $$ = $2; }
  ;

assignment_statement:
    T_IDENTIFIER '=' additive_expression
	  { $$ = newAssignmentStatement( $1, equals_assign, $3 ); }
  | T_IDENTIFIER T_INCR
      { $$ = newAssignmentStatement( $1, increment_assign, newIntegerExpr( integer_expr, 1 ) ); }
  | T_IDENTIFIER T_DECR
      { $$ = newAssignmentStatement( $1, decrement_assign, newIntegerExpr( integer_expr, 1 ) ); }
  | T_IDENTIFIER T_PLUSEQ primary_expression
      { $$ = newAssignmentStatement( $1, plusequals_assign, $3 ); }
  | T_IDENTIFIER T_MINUSEQ primary_expression
      { $$ = newAssignmentStatement( $1, minusequals_assign, $3 ); }
  | T_IDENTIFIER T_TIMESEQ primary_expression
      { $$ = newAssignmentStatement( $1, timesequals_assign, $3 ); }
  | T_IDENTIFIER T_DIVIDEEQ primary_expression
      { $$ = newAssignmentStatement( $1, divideequals_assign, $3 ); }
  ;

opt_else_clause:
    /* empty */
      { $$ = NULL; }
  | T_ELSE statement_block
      { $$ = $2; }
  | T_ELSE T_IF '(' comparison_expression ')' statement_block opt_else_clause
      { $$ = newStatementList( newIfStatement( $4, $6, $7 ), NULL ); }
  ;

case_block:
    case_clause opt_case_clause_list statement opt_statement_list
	  { $$ = newCaseBlock( newExpressionList( $1, $2 ), newStatementList( $3, $4 ) ); }
  ;

opt_case_block_list:
    /* empty */
	  { $$ = NULL; }
  | case_block opt_case_block_list
      { $$ = newCaseBlockList( $1, $2 ); }
  ;

case_clause:
    T_CASE primary_expression ':'
	  { $$ = $2; }
  |
	T_DEFAULT ':'
	  { $$ = newDefaultExpr(); }
  ;

opt_case_clause_list:
    /* empty */
	  { $$ = NULL; }
  | case_clause opt_case_clause_list
      { $$ = newExpressionList( $1, $2 ); }
  ;


/*******************************************************************************
 *                             Expressions                                     *
 ******************************************************************************/

opt_expression_list:
    /* empty */
      { $$ = NULL; }
  | expression_list
      { $$ = $1; }
  ;

expression_list:
    additive_expression expression_list_tail
      { $$ = newExpressionList( $1, $2 ); }
    ;

expression_list_tail:
    /* empty */
      { $$ = NULL; }
  | ',' additive_expression expression_list_tail
      { $$ = newExpressionList( $2, $3 ); }
  ;

comparison_expression:
    T_BOOLEAN
	  { $$ = newIntegerExpr( boolean_expr, $1 ); }
  | primary_expression T_EQUALS primary_expression
      { $$ = newBinaryOpExpression( equality_expr, $1, $3 ); }
  | primary_expression T_NEQ primary_expression
      { $$ = newBinaryOpExpression( inequality_expr, $1, $3 ); }
  | primary_expression '<' primary_expression
      { $$ = newBinaryOpExpression( lessthan_expr, $1, $3 ); }
  | primary_expression '>' primary_expression
      { $$ = newBinaryOpExpression( greaterthan_expr, $1, $3 ); }
  | primary_expression T_LEQ primary_expression
      { $$ = newBinaryOpExpression( lessequal_expr, $1, $3 ); }
  | primary_expression T_GEQ primary_expression
      { $$ = newBinaryOpExpression( greaterequal_expr, $1, $3 ); }
  ;

additive_expression:
    primary_expression
      { $$ = $1; }
  | primary_expression '+' primary_expression
      { $$ = newBinaryOpExpression( addition_expr, $1, $3 ); }
  | primary_expression '-' primary_expression
      { $$ = newBinaryOpExpression( subtraction_expr, $1, $3 ); }
  | primary_expression '*' primary_expression
      { $$ = newBinaryOpExpression( multiplication_expr, $1, $3 ); }
  | primary_expression '/' primary_expression
      { $$ = newBinaryOpExpression( division_expr, $1, $3 ); }
  ;

primary_expression:
    T_INTEGER
      { $$ = newIntegerExpr( integer_expr, $1 ); }
  | T_STRING
      { $$ = newStringExpr( $1 ); }
  | T_IDENTIFIER '(' opt_expression_list ')'
      { $$ = newCallExpression( NULL, $1, $3 ); }
  | T_IDENTIFIER '(' error ')'
      { syntax_error( "bad call expression" ); freeIdent( $1 ); $$ = NULL; }
  | T_IDENTIFIER '.' T_IDENTIFIER '(' opt_expression_list ')'
      { $$ = newCallExpression( $1, $3, $5 ); }
  | T_IDENTIFIER '.' T_IDENTIFIER '(' error ')'
      { syntax_error( "bad call expression" ); freeIdent( $1 ); freeIdent( $3 ); $$ = NULL; }
  | T_IDENTIFIER '.' T_IDENTIFIER
      { $$ = newConstantExpression( $1, $3 ); }
  | T_IDENTIFIER
      { $$ = newVariableExpr( $1 ); }
  ;

%%
