/******************************************************************************
 *
 * Filename: hashtable.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  11th June 2003
 *
 * Purpose:  Java-style hashtable for convenience.
 *           Based on hashtable used in JavaVM.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"

#include "hashtable.h"


/**
 * Java-style hash function.  Calculates a simple hash code for a given string.
 * For long strings, we will actully overflow, but that does not matter.
 */
unsigned long hash( const char* str )
{
	int n, off = 0, len = strlen( str );
	unsigned long ret = 0;

	if( len < 16 )
	{
 	    for( n = len; n > 0; n-- )
	 		ret = (ret * 37) + str[off++];
 	}
	else
	{
 	    int skip = len / 8;
 	    for( n = len ; n > 0; n -= skip, off += skip )
 			ret = (ret * 39) + str[off];
 	}

	return ret;
}


/**
 * Create a new instance of a hashtable with the given number of buckets.
 */
Hashtable* Hashtable_new( int buckets )
{
	int n;

	Hashtable* ret = (Hashtable*) malloc( sizeof(Hashtable) );
	ret->bucketc = buckets;
	ret->bucketv = (Hashtable_Entry**) malloc( buckets * sizeof(Hashtable_Entry*) );
	ret->count   = 0;

	for( n = 0; n < buckets; n++ )
		ret->bucketv[n] = NULL;

	return ret;
}


/**
 * Destroy the given hashtable.  Note that this does not free any keys or
 * values, as it is assumed the caller owns the storage.
 */
void Hashtable_delete( Hashtable* table, size_t valuesize, int keys, int values, DestructorFnPtr destructor )
{
	Hashtable_removeAll( table, valuesize, keys, values, destructor );

	free( table->bucketv );
	table->bucketc = table->count = 0;
	free( table );
}


/**
 * Removes the contents of the given hashtable.  If 'keys' is set, then it
 * frees the keys.  If 'values' is sent, then it frees the values.  If
 * 'destructor' is passed then this is called for each value before it is
 * freed. The hashtable itself is not freed, though the entries hanging from
 * the buckets are.  The hashtable is then empty and ready to be used again.
 */
void Hashtable_removeAll( Hashtable* table, size_t valuesize, int keys, int values, DestructorFnPtr destructor )
{
	Hashtable_Entry *entry, *temp;
	int n;

	/* Loop through all the buckets */
	for( n = 0; n < table->bucketc; n++ )
	{
		/* Follow the linked list for each bucket */
		for( entry = table->bucketv[n]; entry; entry = temp )
		{
			/* If asked, free the key */
			if( keys )
				free( (void*) entry->key );

			entry->key = NULL;

			/* If given a destructor fn, call it on the value */
			if( destructor )
				(*destructor)( entry->value );

			/* If asked, free the value */
			if( values )
				free( entry->value );

			entry->value = NULL;

			/* Remember the next entry */
			temp = entry->next;
			entry->next = NULL;

			/* Free the current entry */
			free( entry );
		}

		table->bucketv[n] = NULL;
	}

	table->count = 0;
}


/**
 * Create a new instance of a hashtable entry.
 */
Hashtable_Entry* newHashtableEntry( Hashtable_Entry* next, const char* key, void* value )
{
	Hashtable_Entry* ret = (Hashtable_Entry*) malloc( sizeof(Hashtable_Entry) );
	ret->next  = next;
	ret->key   = key;
	ret->value = value;
	return ret;
}


/**
 * Increase (or decrease) the number of buckets used by the hashtable in order
 * to reduce (or increase) the length of the linked lists used for each bucket.
 * This is normally used when the loading (ratio of entries to buckets) of the
 * Hashtable gets too high and in order to preserve performance, the table is
 * restructured accordingly.  As re-hashing is quite a slow process, it should
 * not be performed too often.
 */
void Hashtable_rehash( Hashtable* table )
{
	Hashtable_Entry **newbv, *entry, *temp;
	int n, newbucket;
	int buckets = table->bucketc * 2 + 1;

	/* Create a new bucket array with the new number of buckets */
	newbv = (Hashtable_Entry**) malloc( buckets * sizeof(Hashtable_Entry*) );

	for( n = 0; n < buckets; n++ )
		newbv[n] = NULL;

	/* For each entry, calculate its new bucket number and insert it using head
	   insertion.  As there are no duplicate keys we don't need to do any
	   strcmp's. */
	for( n = 0; n < table->bucketc; n++ )
	{
		for( entry = table->bucketv[n]; entry; entry = temp )
		{
			newbucket = hash( entry->key ) % buckets;

			temp = entry->next;
			entry->next = newbv[newbucket];
			newbv[newbucket] = entry;
		}
	}

	/* Free the old bucket array */
	free( table->bucketv );

	/* Put the new bucket array into the table and set the count */
	table->bucketc = buckets;
	table->bucketv = newbv;
}


/**
 * Insert a key/value pair into the given hashtable.  If the key exists, then
 * the existing value is replaced and the previous value returned, otherwise
 * a new entry is created with the given key and value and NULL is returned.
 */
void* Hashtable_put( Hashtable* table, const char* key, void* value )
{
	Hashtable_Entry* entry;
	int bucket;

	if( !key || !value ) {
		printf( "ERROR: Attempt to put NULL key or value to internal hashtable!\n" );
		return NULL;
	}

	/* Find the appropriate bucket */
	bucket = hash( key ) % table->bucketc;

	/* Does the hashtable contain this key? */
	for( entry = table->bucketv[bucket]; entry; entry = entry->next )
	{
		if( strcmp( key, entry->key ) == 0 )
			break;
	}

	/* Did we find the key? */
	if( entry )
	{
		/* Replace the value with our new one */
		Hashtable_Entry* temp = entry->value;
		entry->value = value;
		return temp;
	}
	else
	{
		/* Create a new entry if needed and perform head insertion */
		entry = newHashtableEntry( table->bucketv[bucket], key, value );
		table->bucketv[bucket] = entry;
		table->count++;

		/* If this insert gives l.f. greater than 1 do a re-hash */
		if( table->count > table->bucketc )
			Hashtable_rehash( table );
	}

	return NULL;
}


/**
 * Retrieve the value corresponding to the given key from the specified
 * hashtable.  Returns the value or NULL if the table does not contain the key.
 */
void* Hashtable_get( Hashtable* table, const char* key )
{
	Hashtable_Entry* entry;
	int bucket;

	if( !key ) {
		printf( "ERROR: Attempt to get NULL key from internal hashtable!\n" );
		return NULL;
	}

	/* Find the appropriate bucket */
	bucket = hash( key ) % table->bucketc;

	/* Does the hashtable contain this key? */
	for( entry = table->bucketv[bucket]; entry; entry = entry->next )
	{
		if( strcmp( key, entry->key ) == 0 )
			break;
	}

	if( entry )
		return entry->value;

	return NULL;
}


/**
 * Returns the key if the hashtable contains the given value, null otherwise.
 * Searches the entire set of values to see if 'value' is in the hashtable.
 * This is a sequential search and will, on average, search at least half of
 * the hashtable before finding the value.  Use with caution.
 */
const char* Hashtable_findValue( Hashtable* table, void* value )
{
	Hashtable_Entry* entry;
	int n;

	for( n = 0; n < table->bucketc; n++ )
	{
		for( entry = table->bucketv[n]; entry; entry = entry->next )
		{
			if( value == entry->value )
				return entry->key;
		}
	}

	return NULL;
}
